# requires python3 and rospy to be installed

# ENV["PYTHON"] = "/usr/bin/python3"
# Pkg.build("PyCall")
# using PyCall
# using RobotOS

using Colors
using Caesar
using BSON
using Serialization
using FixedPointNumbers
using NearestNeighbors
# using Manifolds

using Test
using Pkg
using Downloads
using DelimitedFiles

# import Caesar._PCL: FieldMapper, createMapping, PointCloud, PointField, PCLPointCloud2, Header, asType, _PCL_POINTFIELD_FORMAT, FieldMapping, MsgFieldMap, FieldMatches
##

@info "download any necessary test data"

function downloadTestData(datafile, url)
  if 0 === Base.filesize(datafile)
    Base.mkpath(dirname(datafile))
    @info "Downloading $url"
    Downloads.download(url, datafile)
  end
  return datafile
end

testdatafolder = "/tmp/caesar/testdata/"


##
@testset "test Caesar._PCL.PCLPointCloud2 to Caesar._PCL.PointCloud converter." begin
##

radarpclfile = joinpath( testdatafolder,"radar", "convertedRadar", "_PCLPointCloud2_15776.dat")
radarpcl_url = "https://github.com/JuliaRobotics/CaesarTestData.jl/raw/main/data/radar/convertedRadar/_PCLPointCloud2_15776.dat"
downloadTestData(radarpclfile,radarpcl_url)
# testdatafile = joinpath( pkgdir(Caesar), "test", "testdata", "_PCLPointCloud2.bson")
# load presaved test data to test the coverter
# BSON.@load testdatafile PointCloudRef PointCloudTest

## build PCLPointCloud2 to be converted

fid = open(radarpclfile,"r")
data = read(fid)
close(fid)

##

pc2 = Caesar._PCL.PCLPointCloud2(;
  header = Caesar._PCL.Header(;
    seq = 92147,
    stamp = 0x0005b24647c4af10,
    frame_id = "velodyne"
  ),
  height = 1,
  width = 493,
  fields = [
    Caesar._PCL.PointField("x", 0x00000000, Caesar._PCL._PCL_FLOAT32, 0x00000001),
    Caesar._PCL.PointField("y", 0x00000004, Caesar._PCL._PCL_FLOAT32, 0x00000001),
    Caesar._PCL.PointField("z", 0x00000008, Caesar._PCL._PCL_FLOAT32, 0x00000001),
    Caesar._PCL.PointField("intensity", 0x00000010, Caesar._PCL._PCL_FLOAT32, 0x00000001)
  ],
  point_step = 32,
  row_step = 15776,
  is_dense = 1,
  data
)


##

show(pc2)

# test major unpacking / type conversion needed by ROS.sensor_msgs.pointcloud2
pc = Caesar._PCL.PointCloud(pc2)

show(pc)

@test pc.height == 1
@test pc.width == 493
@test pc.is_dense

@test length(pc.points) == 493

##

# truncated copy of reference data on good to 1e-3
@test isapprox( [-0,0,0],              pc.points[1].data[1:3], atol=5e-3)
@test isapprox( [-1.56486,1.09851,0],  pc.points[2].data[1:3], atol=5e-3)
@test isapprox( [-793.383,556.945,0],  pc.points[3].data[1:3], atol=5e-3)
@test isapprox( [-0,0,0],              pc.points[4].data[1:3], atol=5e-3)
@test isapprox( [-1.56148,1.10331,0],  pc.points[5].data[1:3], atol=5e-3)
@test isapprox( [-788.548,557.169,0],  pc.points[6].data[1:3], atol=5e-3)
@test isapprox( [-790.109,558.273,0],  pc.points[7].data[1:3], atol=5e-3)
@test isapprox( [-791.671,559.376,0],  pc.points[8].data[1:3], atol=5e-3)
@test isapprox( [-793.232,560.479,0],  pc.points[9].data[1:3], atol=5e-3)
@test isapprox( [-794.794,561.583,0],  pc.points[10].data[1:3], atol=5e-3)


# for (i,pt) in enumerate(pc.points)
#   @test isapprox( PointCloudRef[1][i,1:3], pt.data[1:3]; atol=1e-6)
# end

##

@info "test `apply` transform SpecialEuclidean(2) to 2D pointcloud."

M = SpecialEuclidean(2)
rPc = ArrayPartition([1.;0], [1 0; 0 1.])

pc_ = Caesar._PCL.apply(M, rPc, pc)


@test isapprox( [1-0,0,0],              pc_.points[1].data[1:3], atol=5e-3)
@test isapprox( [1-1.56486,1.09851,0],  pc_.points[2].data[1:3], atol=5e-3)
@test isapprox( [1-793.383,556.945,0],  pc_.points[3].data[1:3], atol=5e-3)
@test isapprox( [1-0,0,0],              pc_.points[4].data[1:3], atol=5e-3)
@test isapprox( [1-1.56148,1.10331,0],  pc_.points[5].data[1:3], atol=5e-3)
@test isapprox( [1-788.548,557.169,0],  pc_.points[6].data[1:3], atol=5e-3)
@test isapprox( [1-790.109,558.273,0],  pc_.points[7].data[1:3], atol=5e-3)
@test isapprox( [1-791.671,559.376,0],  pc_.points[8].data[1:3], atol=5e-3)
@test isapprox( [1-793.232,560.479,0],  pc_.points[9].data[1:3], atol=5e-3)
@test isapprox( [1-794.794,561.583,0],  pc_.points[10].data[1:3], atol=5e-3)


##
end


if v"1.7" <= VERSION

@testset "PandarXT test point cloud conversion test" begin
##

pandarfile = joinpath(testdatafolder,"lidar","simpleICP","_pandar_PCLPointCloud2.jldat")
pandar_url = "https://github.com/JuliaRobotics/CaesarTestData.jl/raw/main/data/lidar/pandar/_pandar_PCLPointCloud2.jldat"
downloadTestData(pandarfile,pandar_url)


# Alternative approach, see more hardcoded test data example (only .data writen to binary) for _PCLPointCloud2_15776.dat"
@info "Loading testdata/_pandar_PCLPointCloud2.jldat which via `Serialization.serialize` of a `Caesar._PCL.PCLPointCloud2` object, at JL 1.7.3, CJL v0.13.1+" 
pc2 = Serialization.deserialize(pandarfile)
# pc2.fields
#   6-element Vector{Caesar._PCL.PointField}:
#  Caesar._PCL.PointField("x", 0x00000000, Caesar._PCL._PCL_FLOAT32, 0x00000001)
#  Caesar._PCL.PointField("y", 0x00000004, Caesar._PCL._PCL_FLOAT32, 0x00000001)
#  Caesar._PCL.PointField("z", 0x00000008, Caesar._PCL._PCL_FLOAT32, 0x00000001)
#  Caesar._PCL.PointField("intensity", 0x00000010, Caesar._PCL._PCL_FLOAT32, 0x00000001)
#  Caesar._PCL.PointField("timestamp", 0x00000018, Caesar._PCL._PCL_FLOAT64, 0x00000001)
#  Caesar._PCL.PointField("ring", 0x00000020, Caesar._PCL._PCL_UINT16, 0x00000001)

##

pc = Caesar._PCL.PointCloud(pc2)

@test UInt32(1) === pc.height
@test UInt32(58015) === pc.width
@test 58015 === length(pc.points)
@test pc.points[1] isa Caesar._PCL.PointXYZ{RGBA{FixedPointNumbers.N0f8}, Float32}
@test pc.is_dense
@test pc.header.seq === UInt32(24285)

# check actual points
@test isapprox( Float32[0.009075656, 3.714255, 0.99023366, 0.0], pc.points[1].data[1:4], atol=5e-3)
@test isapprox( Float32[0.008430854, 3.715781, 0.92004687, 0.0], pc.points[2].data[1:4], atol=5e-3)
@test isapprox( Float32[0.0084575275, 3.727537, 0.8534482, 0.0], pc.points[3].data[1:4], atol=5e-3)
@test isapprox( Float32[0.008463516, 3.7301762, 0.78552955, 0.0], pc.points[4].data[1:4], atol=5e-3)
@test isapprox( Float32[0.0032057164, 1.3119546, 0.25231096, 0.0], pc.points[5].data[1:4], atol=5e-3)

## convert back to PCLPointCloud2

@info "test back conversion from _PCL.PointCloud to _PCL.PCLPointCloud2, and again _PCL.PointCloud"

_pc2 = Caesar._PCL.PCLPointCloud2(pc)
_pc = Caesar._PCL.PointCloud(_pc2)

@test UInt32(1) === _pc.height
@test UInt32(58015) === _pc.width
@test 58015 === length(_pc.points)
@test _pc.points[1] isa Caesar._PCL.PointXYZ{RGBA{FixedPointNumbers.N0f8}, Float32}
@test _pc.is_dense
@test _pc.header.seq === UInt32(24285)

# check actual points
@test isapprox( Float32[0.009075656, 3.714255, 0.99023366, 0.0], _pc.points[1].data[1:4], atol=5e-3)
@test isapprox( Float32[0.008430854, 3.715781, 0.92004687, 0.0], _pc.points[2].data[1:4], atol=5e-3)
@test isapprox( Float32[0.0084575275, 3.727537, 0.8534482, 0.0], _pc.points[3].data[1:4], atol=5e-3)
@test isapprox( Float32[0.008463516, 3.7301762, 0.78552955, 0.0], _pc.points[4].data[1:4], atol=5e-3)
@test isapprox( Float32[0.0032057164, 1.3119546, 0.25231096, 0.0], _pc.points[5].data[1:4], atol=5e-3)


## test SpecialEuclidean(3) transform application

M = SpecialEuclidean(3)
rPc = ArrayPartition([-1;0;10.], [1 0 0; 0 1 0; 0 0 1.])

pc_3D = Caesar._PCL.apply(M, rPc, pc)


@test UInt32(1) === pc_3D.height
@test UInt32(58015) === pc_3D.width
@test 58015 === length(pc_3D.points)
@test pc_3D.points[1] isa Caesar._PCL.PointXYZ{RGBA{FixedPointNumbers.N0f8}, Float32}
@test pc_3D.is_dense
@test pc_3D.header.seq === UInt32(24285)

@test isapprox( Float32[-1+0.009075656, 3.714255,  10+ 0.99023366, 0.0], pc_3D.points[1].data[1:4], atol=5e-3)
@test isapprox( Float32[-1+0.008430854, 3.715781,  10+ 0.92004687, 0.0], pc_3D.points[2].data[1:4], atol=5e-3)
@test isapprox( Float32[-1+0.0084575275, 3.727537, 10+ 0.8534482, 0.0], pc_3D.points[3].data[1:4], atol=5e-3)
@test isapprox( Float32[-1+0.008463516, 3.7301762, 10+ 0.78552955, 0.0], pc_3D.points[4].data[1:4], atol=5e-3)
@test isapprox( Float32[-1+0.0032057164, 1.3119546,10+ 0.25231096, 0.0], pc_3D.points[5].data[1:4], atol=5e-3)


##
end

else
  @error "Skip PointCloud testdata serialization owning to Julia version."
  @test_broken false
end



@testset "Test PointCloud on Terrestrial Lidar and ICP_Simple alignment" begin
##

lidar_terr1_file = joinpath(testdatafolder,"lidar","simpleICP","terrestrial_lidar1.xyz")
lidar_terr1_url = "https://github.com/JuliaRobotics/CaesarTestData.jl/raw/main/data/lidar/simpleICP/terrestrial_lidar1.xyz"
downloadTestData(lidar_terr1_file,lidar_terr1_url)

lidar_terr2_file = joinpath(testdatafolder,"lidar","simpleICP","terrestrial_lidar2.xyz")
lidar_terr2_url = "https://github.com/JuliaRobotics/CaesarTestData.jl/raw/main/data/lidar/simpleICP/terrestrial_lidar2.xyz"
downloadTestData(lidar_terr2_file,lidar_terr2_url)


# load the data to memory
X_fix = readdlm(lidar_terr1_file, Float32)
X_mov = readdlm(lidar_terr2_file, Float32)

# convert data to PCL types
pc_fix = Caesar._PCL.PointCloud(X_fix);
icp_fix = Caesar._PCL._ICP_PointCloud(pc_fix)

pc_mov = Caesar._PCL.PointCloud(X_mov);
icp_mov = Caesar._PCL._ICP_PointCloud(pc_mov)

@test length(icp_fix) == 1250164

##

neighbors = 10
kdtree = NearestNeighbors.KDTree([icp_fix.x'; icp_fix.y'; icp_fix.z'])
query_points = [icp_fix.x[icp_fix.sel]'; icp_fix.y[icp_fix.sel]'; icp_fix.z[icp_fix.sel]']
idxNN_all_qp, = knn(kdtree, query_points, neighbors, false)

ref_idxNN = [[15, 23, 27, 16, 7, 6, 1, 9, 11, 5], [24, 11, 19, 29, 7, 3, 2, 8, 17, 4], [23, 29, 15, 8, 11, 3, 2, 7, 17, 4], [11, 7, 2, 8, 23, 3, 24, 29, 17, 4], [9, 20, 16, 27, 10, 1, 12, 6, 21, 5]]

@test isapprox(ref_idxNN, idxNN_all_qp[1:5])

##

# estimate normals for patches around each point
Caesar._PCL.estimate_normals!(icp_fix, neighbors)

nx_ref_1_10 = [
 -0.2027982521351298
  0.2695738031305203
 -0.12962918659975647
 -0.18700445755397993
  0.006289904039860751
 -0.06413452235027679
 -0.18425770654086202
  0.4346435000845428
 -0.1495511463188632
  0.15212762938255636
]

ny_ref_1_10 = [
  0.00598286920628624
  0.03450499429584508
  0.00094230477225829
  0.08867086536779206
  0.2020404127780519
  0.1791992164773645
 -0.08868664880709484
 -0.41835406632782735
 -0.28482156811677
 -0.1355238211691936
]

nz_ref_1_10 = [
  0.9792022641962165
  -0.9623613510705796
   0.991562093891856
   0.9783490228389194
   0.9793569873706134
   0.9817201250136417
   0.9788686203488539
  -0.7975367722062914
   0.9468427160675728
   0.9790252694768118 
]

@test_broken isapprox(nx_ref_1_10, icp_fix.nx[1:10])
@test_broken isapprox(ny_ref_1_10, icp_fix.ny[1:10])
@test_broken isapprox(nz_ref_1_10, icp_fix.nz[1:10])
# FIXME, DUPLICATE BAD, BUT ONLY TWO OF 10 POINTS SEEM FLIPPED
@test isapprox(nx_ref_1_10, icp_fix.nx[1:10]; atol = 1.5)
@test isapprox(ny_ref_1_10, icp_fix.ny[1:10]; atol = 1.5)
# @test isapprox(nz_ref_1_10, icp_fix.nz[1:10]; atol = 1.5)


# test select_in_range
max_overlap_distance=Inf
sir = Caesar._PCL.select_in_range!(icp_fix, X_mov, max_overlap_distance)

sir_ref_1_10 = [
  1
  2
  3
  4
  5
  6
  7
  8
  9
 10
]

@test isapprox(sir_ref_1_10, sir[1:10])

##

# must run estimate_normals first
initial_distances = Caesar._PCL.matching!(icp_mov, icp_fix)

mtch_ref_1_10 = [
  0.07132447835334486
  -0.07754577389160945
  0.06451659848545357
  0.06904256092358588
  0.05010243020270855
  0.05711289001579725
  0.06957921459621456
  -0.08329541405485588
  0.06695342406795207
  0.038976789495066375
]

@test_broken isapprox(mtch_ref_1_10, initial_distances[1:10])
# FIXME, duplicate must be removed, 2 of 10 not right
@test isapprox(mtch_ref_1_10, initial_distances[1:10]; atol=0.5)

##

min_planarity = 0.3
Caesar._PCL.reject!(icp_fix, icp_mov, min_planarity, initial_distances)

## Manifolds based rigid transform check

# x = [
#   -1.1362446960310095
#   0.29917409927138205
#  -1.6612742451109666
#   0.7432673613376161
#   0.0998255815132224
#  -1.1510640331762294
# ]

# R_ref = [
#   1.0        1.66127  0.299174
#   -1.66127    1.0      1.13624
#   -0.299174  -1.13624  1.0 
# ]


# M = SpecialEuclidean(3)
# Mr = M.manifold[2]
# R = exp_lie(Mr, hat(Mr, Identity(Mr), x[4:6]))
# fPm = ArrayPartition(x[1:3])

## check alignment

H, pcmovd, stat = Caesar._PCL.alignICP_Simple(X_fix, X_mov)

##

# LieAlgebra expontiation based ref:
H_ref = [
  0.999392    4.89177e-5  -0.0348611   0.0995209
  -0.00126357  0.999393    -0.0348214  -0.0498342
   0.0348382   0.0348443    0.998785   -0.0499514
   0.0         0.0          0.0         1.0
]
# non-rigid (un-normalized R)
# H_ref = [
#   1.00101     5.66661e-5  -0.0349906   0.0994922
#   -0.00119442  0.999613    -0.0344838  -0.0498164
#    0.0350353   0.035119     1.00047    -0.0501982
#    0.0         0.0          0.0         1.0 
# ]

# with rotation Lie Algebra exp map modification
pcmv_ref_10 = [
  0.634187  0.0414235  -0.429596
  0.639184  0.0424166  -0.429387
  0.638185  0.0424179  -0.429422
  0.638185  0.0404191  -0.429491
  0.632188  0.0404267  -0.4297
  0.634187  0.0404241  -0.429631
  0.636186  0.041421   -0.429526
  0.640184  0.0404166  -0.429421
  0.633188  0.0434236  -0.429561
  0.63019   0.042428   -0.4297
]
pcmv_ref_end = [
  -0.440997  0.232817  0.109241
  -0.444994  0.231823  0.109067
  -0.432001  0.249795  0.110147
  -0.433001  0.247798  0.110043
  -0.438997  0.253802  0.110043
  -0.436998  0.256797  0.110217
  -0.434999  0.249799  0.110043
  -0.437997  0.251802  0.110008
  -0.436999  0.237809  0.109555
  -0.446993  0.233824  0.109067
]

# # non-rigid legacy
# pcmv_ref_10 = [
#   0.635052  0.0413591  -0.430395
#   0.640057  0.0423527  -0.430185
#   0.639056  0.0423539  -0.43022
#   0.639055  0.0403547  -0.43029
#   0.633049  0.0403618  -0.4305
#   0.635051  0.0403594  -0.43043
#   0.637054  0.0413567  -0.430325
#   0.641058  0.0403523  -0.43022
#   0.634051  0.0433595  -0.43036
#   0.631048  0.0423634  -0.4305 
# ]
# pcmv_ref_end = [
#   -0.441908  0.232918  0.109252
#   -0.445912  0.231923  0.109076
#   -0.432898  0.2499    0.110164
#   -0.433899  0.247902  0.110059
#   -0.439904  0.253907  0.110059
#   -0.437902  0.256904  0.110235
#   -0.435901  0.249904  0.110059
#   -0.438903  0.251907  0.110024
#   -0.437903  0.237911  0.109567
#   -0.447914  0.233925  0.109076
# ]


@test isapprox(H_ref, H; atol = 1e-4)
@test isapprox(pcmv_ref_10, pcmovd[1:10,:]; atol = 1e-4)
@test isapprox(pcmv_ref_end, pcmovd[end-9:end,:]; atol = 1e-4)


##
end


#